@@ -7,6 +7,7 @@ import haiku as hk
 
 
 class MuZero:
+  r"""Muzero algorithm"""
   def __init__(self, 
                embedding_dim,
                num_actions,
@@ -16,6 +17,25 @@ class MuZero:
                optimizer = optax.adam(0.01),
                discount: float = 0.99
                ):
+    """
+    TODO: more args, more flexible for repr, pred, dy modules
+    
+    Parameters
+    ----------
+    embedding_dim: Any. 
+        The embedding dimension of hidden state `s`. Depends on the representation module. 
+    num_actions: Any. 
+        The maximum number of actions the policy can take. Depends on the prediction module and dynamic module.
+    representation_module: A class inherents hk.Module, 
+        which takes raw observation `obs` from the environment as input and outputs the hidden state `s`.
+        `s` will be the input of prediction_module and dynamic_module.
+    prediction_module: A class inherents hk.Module, 
+        which takes hidden state `s` as input and outputs prior logits `logits` and value `v` of the state.
+    dynamic_module: A class inherents hk.Module,
+        which takes hidden state `s` and action `a` as input and outputs reward `r` and next hidden state `ns`.
+    optimizer: Optimizer to update the parameters of `representation_module`, `prediction_module` and `dynamic_module`.
+    discount: Any. Used for mctx.RecurrentFnOutput.
+    """
     self.repr_func = self._init_representation_func(representation_module, 
                                                     embedding_dim) 
     self.repr_func = hk.without_apply_rng(hk.transform(self.repr_func))
@@ -37,6 +57,17 @@ class MuZero:
   
   def init(self, rng_key, sample_input):
     # TODO doc sample_input shape requirement
+    """Inits `representation`, `prediction` and `dynamic` functions and optimizer
+    
+    Parameters
+    ----------
+    rng_key: jax.random.PRNGKey.
+    sample_input: Array. The dimension is `[B, ...]` where B is the batch dimension.
+    
+    Returns
+    ----------
+    params: dict. {'representation': repr_params, 'prediction': pred_params, 'dynamic': dy_params}
+    """
     repr_params = self.repr_func.init(rng_key, sample_input)
     s = self.repr_func.apply(repr_params, sample_input)
     pred_params = self.pred_func.init(rng_key, s)
@@ -65,6 +96,22 @@ class MuZero:
           obs_from_batch: bool = False,
           num_simulations: int = 5,
           *args, **kwargs):
+    """Acts given environment's observations.
+    
+    Parameters
+    ----------
+    rng_key: jax.random.PRNGKey.
+    obs: Array. The raw observations from environemnt.
+    with_pi: bool.
+    with_value: bool.
+    obs_from_batch: bool.
+    num_simulations: int, positive int. Argument for mctx.muzero_policy
+    args, kwargs: Arguments for mctx.muzero_policy
+    
+    Returns
+    ----------
+    
+    """
     if not obs_from_batch:
       obs = jnp.expand_dims(obs, axis=0)
     plan_output, root_value = self._plan(self.params, rng_key, obs, num_simulations,
@@ -79,8 +126,7 @@ class MuZero:
 
   def train(self, batch, c: float = 1e-4):
     loss, grads = jax.value_and_grad(self._loss_fn)(self._params, batch, c)
-    updates, self._opt_state = self._optimizer.update(grads, self._opt_state)
-    self._params = optax.apply_updates(self._params, updates)
+    self._params, self._opt_state = self._update(self._params, self._opt_state, grads)
     loss_metric = {'loss': loss.item()}
     return loss_metric
 
@@ -100,6 +146,12 @@ class MuZero:
                                num_simulations=num_simulations,
                                *args, **kwargs)
     return plan_output, root.value
+    
+  @partial(jax.jit, static_argnums=(0,))
+  def _update(self, params, optimizer_state, grads):
+    updates, optimizer_state = self._optimizer.update(grads, optimizer_state)
+    params = optax.apply_updates(params, updates)
+    return params, optimizer_state
   
   @partial(jax.jit, static_argnums=(0, 3))
   def _loss_fn(self, params, batch, c: float = 1e-4):
@@ -198,4 +250,4 @@ class MuZero:
     def dynamic_func(s, a):
       dy_model = dynamic_module(embedding_dim, num_actions)
       return dy_model(s, a)
-    return dynamic_func 
\ No newline at end of file
+    return dynamic_func 
